;;; Copyright (C) 2020 Wolfgang Corcoran-Mathe
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a
;;; copy of this software and associated documentation files (the
;;; "Software"), to deal in the Software without restriction, including
;;; without limitation the rights to use, copy, modify, merge, publish,
;;; distribute, sublicense, and/or sell copies of the Software, and to
;;; permit persons to whom the Software is furnished to do so, subject to
;;; the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included
;;; in all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
;;; IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
;;; CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
;;; TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
;;; SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(: exact-natural? (* -> boolean))
(define (exact-natural? x)
  (and (exact-integer? x) (not (negative? x))))

;; Find the least element of a list non-empty of naturals. If an element
;; is zero, returns it immediately.
(: short-minimum ((list-of integer) -> integer))
(define (short-minimum ns)
  (call-with-current-continuation
   (lambda (return)
     (reduce (lambda (n s)
               (if (zero? n) (return n) (min n s)))
             0
             ns))))

(: sum ((list-of integer) -> integer))
(define (sum ns) (reduce + 0 ns))

(define-type range (struct <range>))

(define-record-type <range>
  (raw-range start-index
             length
             indexer
             complexity)
  range?
  (start-index range-start-index : integer)
  (length range-length : integer)
  (indexer range-indexer : (integer -> *))
  (complexity range-complexity : fixnum))

;; Maximum number of indexers to compose with range-reverse and
;; range-append before a range is expanded with vector-range.
;; This may need adjustment.
(: %range-maximum-complexity fixnum)
(define %range-maximum-complexity 16)

;; Returns an empty range which is otherwise identical to r.
(: %empty-range-from (range -> range))
(define (%empty-range-from r)
  (raw-range (range-start-index r)
             0
             (range-indexer r)
             (range-complexity r)))

(: threshold? (fixnum -> boolean))
(define (threshold? k)
  (> k %range-maximum-complexity))

(: %range-valid-index? (range integer -> boolean))
(define (%range-valid-index? r index)
  (and (exact-natural? index)
       (< index (range-length r))))

;; As the previous check, but bound is assumed to be exclusive.
(: %range-valid-bound? (range integer -> boolean))
(define (%range-valid-bound? r bound)
  (and (exact-natural? bound)
       (<= bound (range-length r))))

;;;; Constructors

;; The primary range constructor does some extra consistency checking.
(: range (integer (integer -> *) -> range))
(define (range length indexer)
  (assert 'range (exact-natural? length))
  (assert 'range (procedure? indexer))
  (raw-range 0 length indexer 0))

(: numeric-range (or (integer integer -> range)
                     (integer integer integer -> range)))
(define numeric-range
  (case-lambda
    ((start end) (numeric-range start end 1))
    ((start end step)
     (assert 'numeric-range (real? start))
     (assert 'numeric-range (real? end))
     (assert 'numeric-range (not (zero? step)) "zero-valued step")
     (let ((len (exact (ceiling (max 0 (/ (- end start) step))))))
       ;; Try to ensure that we can compute a correct range from the
       ;; given parameters, i.e. one not plagued by roundoff errors.
       (assert 'numeric-range
               (cond ((and (positive? step) (< start end))
                      (and (> (+ start step) start)
                           (< (+ start (* (- len 1) step)) end)))
                     ((and (negative? step) (> start end))
                      (and (< (+ start step) start)
                           (> (+ start (* (- len 1) step)) end)))
                     (else #t))
               "cannot compute range from parameters")
       (raw-range 0 len (lambda (n) (+ start (* n step))) 0)))))

;; TODO: Consider possible round-off bugs.
(: iota-range (or (integer -> range)
                  (integer integer -> range)
                  (integer integer integer -> range)))
(define iota-range
  (case-lambda
    ((len) (iota-range len 0 1))
    ((len start) (iota-range len start 1))
    ((len start step)
     (assert 'iota-range (exact-natural? len))
     (assert 'iota-range (real? start))
     (assert 'iota-range (real? step))
     (raw-range 0
                len
                (cond ((and (zero? start) (= step 1)) (lambda (i) i))
                      ((= step 1) (lambda (i) (+ start i)))
                      ((zero? start) (lambda (i) (* step i)))
                      (else (lambda (i) (+ start (* step i)))))
                0))))

(: vector-range (vector -> range))
(define (vector-range vec)
  (assert 'vector-range (vector? vec))
  (raw-range 0 (vector-length vec) (lambda (i) (vector-ref vec i)) 0))

;; This implementation assumes that string-ref is O(n), as would be
;; the case with UTF-8.  If an implementation has an O(1) string-ref,
;; the following version is preferable:
;;
;; (raw-range 0 (string-length s) (lambda (i) (string-ref s i))))
;;
(: string-range (string -> range))
(define (string-range s)
  (assert 'string-range (string? s))
  (vector-range (string->vector s)))

(: %range-maybe-vectorize (range -> range))
(define (%range-maybe-vectorize r)
  (if (threshold? (range-complexity r))
      (vector-range (range->vector r))
      r))

;;;; Accessors

(: range-ref (range integer -> *))
(define (range-ref r index)
  (assert 'range-ref (range? r))
  (assert 'range-ref (%range-valid-index? r index) "invalid index")
  ((range-indexer r) (+ index (range-start-index r))))

;; A portable implementation can't rely on inlining, but it
;; can rely on macros.
(define-syntax %range-ref-no-check
  (syntax-rules ()
    ((_ r index)
     ((range-indexer r) (+ index (range-start-index r))))))

(: range-first (range -> *))
(define (range-first r)
  (assert 'range-first (range? r))
  (%range-ref-no-check r (range-start-index r)))

(: range-last (range -> *))
(define (range-last r)
  (assert 'range-last (range? r))
  (%range-ref-no-check r (- (range-length r) 1)))

;;;; Predicates

(: range=? (or ((* * -> boolean) range range -> boolean)
               ((* * -> boolean) #!rest range -> boolean)))
(define range=?
  (case-lambda
    ((equal ra rb)                      ; two-range fast path
     (assert 'range=? (procedure? equal))
     (assert 'range=? (range? ra))
     (%range=?-2 equal ra rb))
    ((equal . rs)                       ; variadic path
     (assert 'range=? (procedure? equal))
     (assert 'range=? (pair? rs) "invalid number of arguments")
     (let ((ra (car rs)))
       (assert 'range=? (range? ra))
       (every (lambda (rb) (%range=?-2 equal ra rb)) (cdr rs))))))

(: %range=?-2 ((* * -> boolean) range range -> boolean))
(define (%range=?-2 equal ra rb)
  (assert 'range=? (range? rb))
  (or (eqv? ra rb)                      ; quick check
      (let ((la (range-length ra)))
        (and (= la (range-length rb))
             (if (zero? la)
                 #t                     ; all empty ranges are equal
                 (let lp ((i 0))
                   (cond ((= i la) #t)
                         ((not (equal (%range-ref-no-check ra i)
                                      (%range-ref-no-check rb i)))
                          #f)
                         (else (lp (+ i 1))))))))))

;;;; Iteration

(: range-split-at (range integer -> range range))
(define (range-split-at r index)
  (assert 'range-split-at (range? r))
  (assert 'range-split-at (%range-valid-bound? r index) "invalid index")
  (cond ((= index 0) (values (%empty-range-from r) r))
        ((= index (range-length r)) (values r (%empty-range-from r)))
        (else
         (let ((indexer (range-indexer r)) (k (range-complexity r)))
           (values (raw-range (range-start-index r) index indexer k)
                   (raw-range index
                              (- (range-length r) index)
                              indexer
                              k))))))

(: subrange (range integer integer -> range))
(define (subrange r start end)
  (assert 'subrange (range? r))
  (assert 'subrange
          (%range-valid-index? r start)
          "invalid start index argument")
  (assert 'subrange
          (%range-valid-bound? r end)
          "invalid end index argument")
  (assert 'subrange
          (not (negative? (- end start)))
          "invalid subrange")
  (if (and (zero? start) (= end (range-length r)))
      r
      (raw-range (+ (range-start-index r) start)
                 (- end start)
                 (range-indexer r)
                 (range-complexity r))))

(: range-segment (range integer -> (list-of range)))
(define (range-segment r k)
  (assert 'range-segment (range? r))
  (assert 'range-segment (and (exact-integer? k) (positive? k)))
  (let ((len (range-length r))
        (%subrange-no-check
         (lambda (s e)
           (raw-range (+ (range-start-index r) s)
                      (- e s)
                      (range-indexer r)
                      (range-complexity r)))))
    (unfold (lambda (i) (>= i len))
            (lambda (i) (%subrange-no-check i (min len (+ i k))))
            (lambda (i) (+ i k))
            0)))

(: range-take (range integer -> range))
(define (range-take r count)
  (assert 'range-take (range? r))
  (assert 'range-take
          (%range-valid-bound? r count)
          "invalid count argument")
  (cond ((zero? count) (%empty-range-from r))
        ((= count (range-length r)) r)
        (else (raw-range (range-start-index r)
                         count
                         (range-indexer r)
                         (range-complexity r)))))

(: range-take-right (range integer -> range))
(define (range-take-right r count)
  (assert 'range-take-right (range? r))
  (assert 'range-take-right
          (%range-valid-bound? r count)
          "invalid count argument")
  (cond ((zero? count) (%empty-range-from r))
        ((= count (range-length r)) r)
        (else
         (raw-range (+ (range-start-index r) (- (range-length r) count))
                    count
                    (range-indexer r)
                    (range-complexity r)))))

(: range-drop (range integer -> range))
(define (range-drop r count)
  (assert 'range-drop (range? r))
  (assert 'range-drop
          (%range-valid-bound? r count)
          "invalid count argument")
  (if (zero? count)
      r
      (raw-range (+ (range-start-index r) count)
                 (- (range-length r) count)
                 (range-indexer r)
                 (range-complexity r))))

(: range-drop-right (range integer -> range))
(define (range-drop-right r count)
  (assert 'range-drop-right (range? r))
  (assert 'range-drop-right
          (%range-valid-bound? r count)
          "invalid count argument")
  (if (zero? count)
      r
      (raw-range (range-start-index r)
                 (- (range-length r) count)
                 (range-indexer r)
                 (range-complexity r))))

(: range-count (procedure range #!rest range -> integer))
(define (range-count pred r . rs)
  (assert 'range-count (procedure? pred))
  (assert 'range-count (range? r))
  (if (null? rs)                        ; one-range fast path
      (%range-fold-1 (lambda (c x)
                       (if (pred x) (+ c 1) c))
                     0
                     r
                     'range-count)
      (apply range-fold                 ; variadic path
             (lambda (c . xs)
               (if (apply pred xs) (+ c 1) c))
             0
             r
             rs)))

(: range-any (procedure range #!rest range -> boolean))
(define (range-any pred r . rs)
  (assert 'range-any (procedure? pred))
  (assert 'range-any (range? r))
  (if (null? rs)                        ; one-range fast path
      (%range-fold-1 (lambda (last x)
                       (or (pred x) last))
                     #f
                     r
                     'range-any)
      (apply range-fold                 ; variadic path
             (lambda (last . xs) (or (apply pred xs) last))
             #f
             r
             rs)))

(: range-every (procedure range #!rest range -> boolean))
(define (range-every pred r . rs)
  (assert 'range-every (procedure? pred))
  (assert 'range-every (range? r))
  (call-with-current-continuation
   (lambda (return)
     (if (null? rs)                     ; one-range fast path
         (%range-fold-1 (lambda (_ x) (or (pred x) (return #f)))
                        #t
                        r
                        'range-every)
         (apply range-fold              ; variadic path
                (lambda (_ . xs) (or (apply pred xs) (return #f)))
                #t
                r
                rs)))))

(: range-map (procedure #!rest range -> range))
(define (range-map proc . rs)
  (assert 'range-map (pair? rs) "invalid number of arguments")
  (vector-range (apply range-map->vector proc rs)))

(: range-filter-map (procedure #!rest range -> range))
(define (range-filter-map proc . rs)
  (assert 'range-filter-map (pair? rs) "invalid number of arguments")
  (vector-range (list->vector (apply range-filter-map->list proc rs))))

(: range-map->list (procedure range #!rest range -> list))
(define (range-map->list proc r . rs)
  (assert 'range-map->list (procedure? proc))
  (if (null? rs)                        ; one-range fast path
      (%range-fold-right-1 (lambda (res x) (cons (proc x) res))
                           '()
                           r
                           'range-map->list)
      (apply range-fold-right           ; variadic path
             (lambda (res . xs) (cons (apply proc xs) res))
             '()
             r
             rs)))

(: range-filter-map->list
   (procedure range #!rest range -> list))
(define (range-filter-map->list proc r . rs)
  (if (null? rs)                        ; one-range fast path
      (%range-fold-right-1 (lambda (res x)
                             (cond ((proc x) =>
                                    (lambda (elt) (cons elt res)))
                                   (else res)))
                           '()
                           r
                           'range-filter-map->list)
      (apply range-fold-right           ; variadic path
             (lambda (res . xs)
               (cond ((apply proc xs) => (lambda (elt) (cons elt res)))
                     (else res)))
             '()
             r
             rs)))

(: range-map->vector (procedure range #!rest range -> vector))
(define (range-map->vector proc r . rs)
  (assert 'range-map->vector (procedure? proc))
  (assert 'range-map->vector (range? r))
  (if (null? rs)                        ; one-range fast path
      (vector-unfold (lambda (i) (proc (%range-ref-no-check r i)))
                     (range-length r))
      (let ((rs* (cons r rs)))          ; variadic path
        (vector-unfold (lambda (i)
                         (apply proc (map (lambda (r)
                                            (%range-ref-no-check r i))
                                          rs*)))
                       (short-minimum (map range-length rs*))))))

(: range-for-each
   (procedure range #!rest range -> undefined))
(define (range-for-each proc r . rs)
  (assert 'range-for-each (procedure? proc))
  (assert 'range-for-each (range? r))
  (if (null? rs)                        ; one-range fast path
      (let ((len (range-length r)))
        (let lp ((i 0))
          (cond ((= i len) (if #f #f))
                (else (proc (%range-ref-no-check r i))
                      (lp (+ i 1))))))
      (let* ((rs* (cons r rs))          ; variadic path
             (len (short-minimum (map range-length rs*))))
        (let lp ((i 0))
          (cond ((= i len) (if #f #f))
                (else
                 (apply proc (map (lambda (r)
                                    (%range-ref-no-check r i))
                                  rs*))
                 (lp (+ i 1))))))))

(: %range-fold-1 (procedure * range symbol -> *))
(define (%range-fold-1 proc nil r loc)
  (assert loc (procedure? proc))
  (assert loc (range? r))
  (let ((len (range-length r)))
    (let lp ((i 0) (acc nil))
      (if (= i len)
          acc
          (lp (+ i 1) (proc acc (%range-ref-no-check r i)))))))

(: range-fold (or (procedure * range -> *)
                  (procedure * #!rest range -> *)))
(define range-fold
  (case-lambda
    ((proc nil r)                       ; one-range fast path
     (%range-fold-1 proc nil r 'range-fold))
    ((proc nil . rs)                    ; variadic path
     (assert 'range-fold (procedure? proc))
     (assert 'range-fold (pair? rs))
     (let ((len (short-minimum (map range-length rs))))
       (let lp ((i 0) (acc nil))
         (if (= i len)
             acc
             (lp (+ i 1)
                 (apply proc acc (map (lambda (r)
                                        (%range-ref-no-check r i))
                                      rs)))))))))

(: %range-fold-right-1 (procedure * range symbol -> *))
(define (%range-fold-right-1 proc nil r loc)
  (assert loc (procedure? proc))
  (assert loc (range? r))
  (let ((len (range-length r)))
    (let rec ((i 0))
      (if (= i len)
          nil
          (proc (rec (+ i 1)) (%range-ref-no-check r i))))))

(: range-fold-right (or (procedure * range -> *)
                        (procedure * #!rest range -> *)))
(define range-fold-right
  (case-lambda
    ((proc nil r)                       ; one-range fast path
     (%range-fold-right-1 proc nil r 'range-fold-right))
    ((proc nil . rs)                    ; variadic path
     (assert 'range-fold-right (procedure? proc))
     (assert 'range-fold-right (pair? rs))
     (let ((len (short-minimum (map range-length rs))))
       (let rec ((i 0))
         (if (= i len)
             nil
             (apply proc
                    (rec (+ i 1))
                    (map (lambda (r) (%range-ref-no-check r i)) rs))))))))

(: range-filter (procedure range -> range))
(define (range-filter pred r)
  (vector-range (list->vector (range-filter->list pred r))))

(: range-filter->list (procedure range -> list))
(define (range-filter->list pred r)
  (assert 'range-filter->list (procedure? pred))
  (assert 'range-filter->list (range? r))
  (%range-fold-right-1 (lambda (xs x)
                         (if (pred x) (cons x xs) xs))
                       '()
                       r
                       'range-filter->list))

(: range-remove (procedure range -> range))
(define (range-remove pred r)
  (vector-range (list->vector (range-remove->list pred r))))

(: range-remove->list (procedure range -> list))
(define (range-remove->list pred r)
  (assert 'range-remove->list (procedure? pred))
  (assert 'range-remove->list (range? r))
  (%range-fold-right-1 (lambda (xs x)
                         (if (pred x) xs (cons x xs)))
                       '()
                       r
                       'range-remove->list))

(: range-reverse (range -> range))
(define (range-reverse r)
  (assert 'range-reverse (range? r))
  (%range-maybe-vectorize
   (raw-range (range-start-index r)
              (range-length r)
              (lambda (n)
                ((range-indexer r) (- (range-length r) 1 n)))
              (+ 1 (range-complexity r)))))

(: range-append (or (-> range)
                    (range -> range)
                    (range range -> range)
                    (#!rest range -> range)))
(define range-append
  (case-lambda
    (() (raw-range 0 0 identity 0))
    ((r) r)                             ; one-range fast path
    ((ra rb)                            ; two-range fast path
     (assert 'range-append (range? ra))
     (assert 'range-append (range? rb))
     (let ((la (range-length ra))
           (lb (range-length rb)))
       (%range-maybe-vectorize          ; FIXME: should be lazy.
        (raw-range 0
                   (+ la lb)
                   (lambda (i)
                     (if (< i la)
                         (%range-ref-no-check ra i)
                         (%range-ref-no-check rb (- i la))))
                   (+ 2 (range-complexity ra) (range-complexity rb))))))
    (rs                                 ; variadic path
     (assert 'range-append (every range? rs))
     (let ((lens (map range-length rs)))
       (%range-maybe-vectorize          ; FIXME: should be lazy.
        (raw-range 0
                   (sum lens)
                   (lambda (i)
                     (let lp ((i i) (rs rs) (lens lens))
                       (if (< i (car lens))
                           (%range-ref-no-check (car rs) i)
                           (lp (- i (car lens)) (cdr rs) (cdr lens)))))
                   (+ (length rs) (sum (map range-complexity rs)))))))))

;;;; Searching

(: range-index
   (procedure range #!rest range -> (or integer false)))
(define (range-index pred r . rs)
  (assert 'range-index (procedure? pred))
  (assert 'range-index (range? r))
  (if (null? rs)                        ; one-range fast path
      (let ((len (range-length r)))
        (let lp ((i 0))
          (cond ((= i len) #f)
                ((pred (%range-ref-no-check r i)) i)
                (else (lp (+ i 1))))))
      (let* ((rs* (cons r rs))          ; variadic path
             (len (short-minimum (map range-length rs*))))
        (let lp ((i 0))
          (cond ((= i len) #f)
                ((apply pred
                        (map (lambda (s) (%range-ref-no-check s i))
                             rs*))
                 i)
                (else (lp (+ i 1))))))))

(: range-index-right
   (procedure range #!rest range -> (or integer false)))
(define (range-index-right pred r . rs)
  (assert 'range-index-right (procedure? pred))
  (assert 'range-index-right (range? r))
  (if (null? rs)                        ; one-range fast path
      (let lp ((i (- (range-length r) 1)))
        (cond ((< i 0) #f)
              ((pred (%range-ref-no-check r i)) i)
              (else (lp (- i 1)))))
      (let ((len (range-length r))      ; variadic path
            (rs* (cons r rs)))
        (assert 'range-index-right
                 (every (lambda (s) (= len (range-length s))) rs)
                "range arguments must be of the same length")
        (let lp ((i (- len 1)))
          (cond ((< i 0) #f)
                ((apply pred
                        (map (lambda (s) (%range-ref-no-check s i))
                             rs*))
                 i)
                (else (lp (- i 1))))))))

(: range-take-while (procedure range -> range))
(define (range-take-while pred r)
  (assert 'range-take-while (procedure? pred))
  (assert 'range-take-while (range? r))
  (cond ((range-index (lambda (x) (not (pred x))) r) =>
         (lambda (i) (range-take r i)))
        (else r)))

(: range-take-while-right (procedure range -> range))
(define (range-take-while-right pred r)
  (assert 'range-take-while-right (procedure? pred))
  (assert 'range-take-while-right (range? r))
  (cond ((range-index-right (lambda (x) (not (pred x))) r) =>
         (lambda (i) (range-take-right r (- (range-length r) 1 i))))
        (else r)))

(: range-drop-while (procedure range -> range))
(define (range-drop-while pred r)
  (assert 'range-drop-while (procedure? pred))
  (assert 'range-drop-while (range? r))
  (cond ((range-index (lambda (x) (not (pred x))) r) =>
         (lambda (i) (range-drop r i)))
        (else (%empty-range-from r))))

(: range-drop-while-right (procedure range -> range))
(define (range-drop-while-right pred r)
  (assert 'range-drop-while-right (procedure? pred))
  (assert 'range-drop-while-right (range? r))
  (cond ((range-index-right (lambda (x) (not (pred x))) r) =>
         (lambda (i) (range-drop-right r (- (range-length r) 1 i))))
        (else (%empty-range-from r))))

;;;; Conversion

(: range->list (range -> list))
(define (range->list r)
  (assert 'range->list (range? r))
  (%range-fold-right-1 xcons '() r 'range->list))

(: range->vector (range -> vector))
(define (range->vector r)
  (assert 'range->vector (range? r))
  (vector-unfold (lambda (i) (%range-ref-no-check r i))
                 (range-length r)))

(: range->string (range -> string))
(define (range->string r)
  (assert 'range->string (range? r))
  (let ((res (make-string (range-length r))))
    (range-fold (lambda (i c) (string-set! res i c) (+ i 1)) 0 r)
    res))

(: vector->range (vector -> range))
(define (vector->range vec)
  (assert 'vector->range (vector? vec))
  (vector-range (vector-copy vec)))

(: range->generator (range -> procedure))
(define (range->generator r)
  (assert 'range->generator (range? r))
  (let ((i 0) (len (range-length r)))
    (lambda ()
      (if (>= i len)
          (eof-object)
          (begin
           (let ((v (%range-ref-no-check r i)))
             (set! i (+ i 1))
             v))))))
